import pathlib
import shutil
from typing import Optional

import typer

from robox import console
from robox.box import presets


def create(
    name: str, preset: Optional[str] = None, path: Optional[pathlib.Path] = None
):
    console.console.print(f'Creating new problem [item]{name}[/item]...')

    preset = preset or 'default'
    preset_cfg = presets.get_installed_preset(preset)

    problem_path = (
        presets.get_preset_installation_path(preset) / preset_cfg.problem
        if preset_cfg.problem is not None
        else presets.get_preset_installation_path('default') / 'problem'
    )

    if not problem_path.is_dir():
        console.console.print(
            f'[error]Problem template [item]{problem_path}[/item] does not exist.[/error]'
        )
        raise typer.Exit(1)

    dest_path = path or pathlib.Path(name)

    if dest_path.exists():
        console.console.print(
            f'[error]Directory [item]{dest_path}[/item] already exists.[/error]'
        )
        raise typer.Exit(1)

    shutil.copytree(str(problem_path), str(dest_path))
    shutil.rmtree(str(dest_path / 'build'), ignore_errors=True)
    shutil.rmtree(str(dest_path / '.box'), ignore_errors=True)



import pathlib
import shlex
import shutil
from pathlib import PosixPath
from typing import Dict, List, Optional, Set

import typer

from robox import console
from robox.box import checkers, package
from robox.box.code import compile_item, run_item
from robox.box.environment import (
    EnvironmentSandbox,
    ExecutionConfig,
)
from robox.box.schema import CodeItem, Generator, Testcase, TestcaseGroup
from robox.box.testcases import find_built_testcases
from robox.grading.judge.cacher import FileCacher
from robox.grading.steps import (
    DigestHolder,
    DigestOrDest,
    DigestOrSource,
)
from robox.utils import StatusProgress


def _compile_generator(generator: CodeItem) -> str:
    return compile_item(generator)


def _get_group_input(group_path: pathlib.Path, i: int) -> pathlib.Path:
    return group_path / f'{i:03d}.in'


def _get_group_output(group_path: pathlib.Path, i: int) -> pathlib.Path:
    return group_path / f'{i:03d}.out'


def _copy_testcase_over(testcase: Testcase, group_path: pathlib.Path, i: int):
    shutil.copy(
        str(testcase.inputPath),
        _get_group_input(group_path, i),
    )
    if testcase.outputPath is not None and testcase.outputPath.is_file():
        shutil.copy(
            str(testcase.outputPath),
            _get_group_output(group_path, i),
        )


def _run_generator(
    generator: Generator,
    args: Optional[str],
    compiled_digest: str,
    group_path: pathlib.Path,
    i: int = 0,
):
    run_log = run_item(
        generator,
        DigestOrSource.create(compiled_digest),
        stdout=DigestOrDest.create(_get_group_input(group_path, i)),
        extra_args=args or None,
    )

    if not run_log or run_log.exitcode != 0:
        console.console.print(
            f'Failed generating test {i} from group path {group_path}',
            style='error',
        )
        raise typer.Exit(1)


def get_all_built_testcases() -> Dict[str, List[Testcase]]:
    pkg = package.find_problem_package_or_die()
    res = {group.name: find_built_testcases(group) for group in pkg.testcases}
    return res


def generate_outputs_for_testcases(
    progress: Optional[StatusProgress] = None, groups: Optional[Set[str]] = None
):
    def step():
        if progress is not None:
            progress.step()

    pkg = package.find_problem_package_or_die()

    built_testcases = get_all_built_testcases()
    main_solution = package.get_main_solution()
    solution_digest: Optional[str] = None

    if main_solution is not None:
        if progress:
            progress.update('Compiling main solution...')
        try:
            solution_digest = compile_item(main_solution)
        except:
            console.console.print('[error]Failed compiling main solution.[/error]')
            raise

    sandbox = EnvironmentSandbox()
    sandbox.timeLimit = pkg.timeLimit * 2
    sandbox.wallTimeLimit = pkg.timeLimit * 2
    sandbox.memoryLimit = pkg.memoryLimit
    extra_config = ExecutionConfig(sandbox=sandbox)

    for group in pkg.testcases:
        if groups is not None and group.name not in groups:
            continue
        group_testcases = built_testcases[group.name]

        for testcase in group_testcases:
            input_path = testcase.inputPath
            output_path = testcase.outputPath

            assert output_path is not None
            if main_solution is None or solution_digest is None:
                console.console.print(
                    'No main solution found to generate outputs for testcases',
                    style='error',
                )
                raise typer.Exit(1)

            try:
                run_log = run_item(
                    main_solution,
                    DigestOrSource.create(solution_digest),
                    stdin=DigestOrSource.create(input_path),
                    stdout=DigestOrDest.create(output_path),
                    extra_config=extra_config,
                )
            except:
                console.console.print(
                    '[error]Failed running main solution to generate testcase.[/error]'
                )
                raise

            if run_log is None or run_log.exitcode != 0:
                console.console.print(
                    f'[error]Failed generating output for [item]{input_path}[/item][/error]',
                )
                if run_log is not None:
                    console.console.print(
                        f'[error]Main solution exited with code [item]{-run_log.exitcode}[/item][/error]',
                    )
                    checker_result = checkers.check_with_no_output(run_log)
                    console.console.print(
                        f'[warning]Time: [item]{run_log.time:.2f}s[/item][/warning]',
                    )
                    console.console.print(
                        f'[warning]Verdict: {checker_result.outcome.value}[/warning]',
                    )
                    console.console.print(
                        f'[warning]Message: [info]{checker_result.message}[/info][/warning]',
                    )
                raise typer.Exit(1)

            step()


def run_generator_script(testcase: TestcaseGroup, cacher: FileCacher) -> str:
    assert testcase.generatorScript is not None
    script_digest = DigestHolder()
    if testcase.generatorScript.path.suffix == '.txt':
        script_digest.value = cacher.put_file_from_path(testcase.generatorScript.path)
    else:
        try:
            compiled_digest = compile_item(testcase.generatorScript)
        except:
            console.console.print(
                f'[error]Failed compiling generator script for group [item]{testcase.name}[/item].[/error]'
            )
            raise

        run_log = run_item(
            testcase.generatorScript,
            DigestOrSource.create(compiled_digest),
            stdout=DigestOrDest.create(script_digest),
        )

        if run_log is None or run_log.exitcode != 0:
            console.console.print(
                f'Could not run generator script for group {testcase.name}'
            )
            raise typer.Exit(1)

    assert script_digest.value
    script = cacher.get_file_content(script_digest.value).decode()
    return script


def extract_script_lines(script: str):
    lines = script.splitlines()
    for line in lines:
        line = line.strip()
        if not line:
            continue

        yield shlex.split(line)[0], shlex.join(shlex.split(line)[1:])


def get_necessary_generators(groups: Set[str], cacher: FileCacher) -> Set[str]:
    pkg = package.find_problem_package_or_die()
    existing_generators = set(generator.name for generator in pkg.generators)

    necessary_generators = set()
    for group in pkg.testcases:
        if groups is not None and group.name not in groups:
            continue

        for generator_call in group.generators:
            necessary_generators.add(generator_call.name)

        if group.generatorScript is not None:
            script = run_generator_script(group, cacher)
            for generator_name, _ in extract_script_lines(script):
                necessary_generators.add(generator_name)

    return existing_generators.intersection(necessary_generators)


def compile_generators(
    progress: Optional[StatusProgress] = None,
    tracked_generators: Optional[Set[str]] = None,
) -> Dict[str, str]:
    def update_status(text: str):
        if progress is not None:
            progress.update(text)

    pkg = package.find_problem_package_or_die()

    generator_to_compiled_digest = {}

    for generator in pkg.generators:
        if tracked_generators is not None and generator.name not in tracked_generators:
            continue
        update_status(f'Compiling generator [item]{generator.name}[/item]')
        try:
            generator_to_compiled_digest[generator.name] = _compile_generator(generator)
        except:
            console.console.print(
                f'[error]Failed compiling generator [item]{generator.name}[/item].[/error]'
            )
            raise

    return generator_to_compiled_digest


def generate_testcases(
    progress: Optional[StatusProgress] = None, groups: Optional[Set[str]] = None
):
    def step():
        if progress is not None:
            progress.step()

    pkg = package.find_problem_package_or_die()
    cacher = package.get_file_cacher()

    compiled_generators = compile_generators(
        progress=progress,
        tracked_generators=get_necessary_generators(groups, cacher)
        if groups is not None
        else None,
    )

    for testcase in pkg.testcases:
        if groups is not None and testcase.name not in groups:
            continue
        group_path = package.get_build_testgroup_path(testcase.name)

        i = 0
        # Individual testcases.
        for tc in testcase.testcases or []:
            _copy_testcase_over(tc, group_path, i)
            i += 1
            step()

        # Glob testcases.
        if testcase.testcaseGlob:
            matched_inputs = sorted(PosixPath().glob(testcase.testcaseGlob))

            for input_path in matched_inputs:
                if not input_path.is_file() or input_path.suffix != '.in':
                    continue
                output_path = input_path.parent / f'{input_path.stem}.out'
                tc = Testcase(inputPath=input_path, outputPath=output_path)
                _copy_testcase_over(tc, group_path, i)
                i += 1
                step()

        # Run single generators.
        for generator_call in testcase.generators:
            generator = package.get_generator(generator_call.name)
            if generator.name not in compiled_generators:
                console.console.print(f'Generator {generator.name} not compiled')
                raise typer.Exit(1)

            _run_generator(
                generator,
                generator_call.args,
                compiled_generators[generator.name],
                group_path,
                i,
            )
            i += 1
            step()

        # Run generator script.
        if testcase.generatorScript is not None:
            script = run_generator_script(testcase, cacher)

            # Run each line from generator script.
            for generator_name, args in extract_script_lines(script):
                generator = package.get_generator(generator_name)
                if generator.name not in compiled_generators:
                    console.console.print(f'Generator {generator.name} not compiled')
                    raise typer.Exit(1)

                _run_generator(
                    generator,
                    args,
                    compiled_generators[generator.name],
                    group_path,
                    i,
                )
                i += 1
                step()



import collections
import pathlib
from typing import Any, Dict, List, Optional, Set

import rich
import rich.table

from robox import console
from robox.box import checkers, environment, package
from robox.box.code import compile_item, run_item
from robox.box.environment import EnvironmentSandbox, ExecutionConfig, VerificationLevel
from robox.box.schema import Solution
from robox.box.testcases import find_built_testcases
from robox.grading.steps import (
    DigestOrDest,
    DigestOrSource,
    Evaluation,
    Outcome,
    TestcaseIO,
    TestcaseLog,
)
from robox.utils import StatusProgress, model_to_yaml


def is_fast(solution: Solution) -> bool:
    # If solution has TLE tag, it is considered slow.
    return not solution.outcome.match(Outcome.TIME_LIMIT_EXCEEDED)


def compile_solutions(
    progress: Optional[StatusProgress] = None,
    tracked_solutions: Optional[Set[str]] = None,
) -> Dict[pathlib.Path, str]:
    pkg = package.find_problem_package_or_die()

    compiled_solutions = {}

    for solution in pkg.solutions:
        if (
            tracked_solutions is not None
            and str(solution.path) not in tracked_solutions
        ):
            continue
        if progress:
            progress.update(f'Compiling solution [item]{solution.path}[/item]...')
        try:
            compiled_solutions[solution.path] = compile_item(solution)
        except:
            console.console.print(
                f'[error]Failed compiling solution [item]{solution.path}[/item].[/error]'
            )
            raise

    return compiled_solutions


def run_solution(
    solution: Solution,
    compiled_digest: str,
    checker_digest: str,
    index: int,
    progress: Optional[StatusProgress] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
) -> Dict[str, List[Evaluation]]:
    pkg = package.find_problem_package_or_die()

    actual_sandbox = package.get_singleton_sandbox()

    sandbox = EnvironmentSandbox()
    sandbox.timeLimit = pkg.timeLimit
    if verification.value >= VerificationLevel.FULL.value:
        # Use double TL.
        sandbox.timeLimit = sandbox.timeLimit * 2
    sandbox.wallTimeLimit = (
        pkg.timeLimit * 2 if actual_sandbox.use_soft_timeout() else sandbox.timeLimit
    )
    sandbox.memoryLimit = pkg.memoryLimit
    extra_config = ExecutionConfig(sandbox=sandbox)

    res = collections.defaultdict(list)

    for group in pkg.testcases:
        testcases = find_built_testcases(group)
        for i, testcase in enumerate(testcases):
            runs_dir = package.get_problem_runs_dir()
            assert testcase.outputPath is not None
            output_path = runs_dir / f'{index}' / group.name / testcase.outputPath.name
            error_path = output_path.with_suffix('.err')
            log_path = output_path.with_suffix('.log')
            output_path.parent.mkdir(parents=True, exist_ok=True)

            if progress:
                progress.update(
                    f'Running solution [item]{solution.path}[/item] on test [item]{group.name}[/item] / [item]{i}[/item]...'
                )
            run_log = run_item(
                solution,
                DigestOrSource.create(compiled_digest),
                stdin=DigestOrSource.create(testcase.inputPath),
                stdout=DigestOrDest.create(output_path),
                stderr=DigestOrDest.create(error_path),
                extra_config=extra_config,
            )

            checker_result = checkers.check(
                checker_digest,
                run_log,
                testcase,
                program_output=output_path,
            )

            eval = Evaluation(
                result=checker_result,
                testcase=TestcaseIO(
                    index=i, input=testcase.inputPath, output=testcase.outputPath
                ),
                log=TestcaseLog(
                    **(run_log.model_dump() if run_log is not None else {}),
                    stdout_absolute_path=output_path.absolute(),
                    stderr_absolute_path=error_path.absolute(),
                    log_absolute_path=log_path.absolute(),
                ),
            )

            log_path.write_text(model_to_yaml(eval))

            res[group.name].append(eval)

    return dict(res)


def run_solutions(
    progress: Optional[StatusProgress] = None,
    tracked_solutions: Optional[Set[str]] = None,
    verification: VerificationLevel = VerificationLevel.NONE,
) -> List[Dict[str, List[Evaluation]]]:
    pkg = package.find_problem_package_or_die()

    checker_digest = checkers.compile_checker()
    compiled_solutions = compile_solutions(
        progress=progress, tracked_solutions=tracked_solutions
    )
    res = []

    for i, solution in enumerate(pkg.solutions):
        if (
            tracked_solutions is not None
            and str(solution.path) not in tracked_solutions
        ):
            res.append({})
            continue
        results_per_group = run_solution(
            solution,
            compiled_solutions[solution.path],
            checker_digest,
            i,
            progress=progress,
            verification=verification,
        )
        res.append(results_per_group)

    return res


def get_outcome_style_verdict(outcome: Outcome) -> str:
    if outcome == Outcome.ACCEPTED:
        return 'green'
    if outcome == Outcome.WRONG_ANSWER:
        return 'red'
    if outcome == Outcome.TIME_LIMIT_EXCEEDED:
        return 'yellow'
    if outcome == Outcome.RUNTIME_ERROR:
        return 'lnumber'
    if outcome == Outcome.MEMORY_LIMIT_EXCEEDED:
        return 'cyan'
    return 'magenta'


def _get_testcase_markup_verdict(eval: Evaluation) -> str:
    res = '✓'
    if eval.result.outcome != Outcome.ACCEPTED:
        res = '✗'
    if eval.result.outcome == Outcome.TIME_LIMIT_EXCEEDED:
        res = '⧖'
    if eval.result.outcome == Outcome.RUNTIME_ERROR:
        res = '✗'
    style = get_outcome_style_verdict(eval.result.outcome)
    res = f'[{style}]{res}[/{style}]'
    if eval.log.stdout_absolute_path:
        output_path = eval.log.stdout_absolute_path.resolve()
        output_link = f'file://{output_path}'
        res = f'[link={output_link}]{res}[/link]'
    return res


def _get_evals_time_in_ms(evals: List[Evaluation]) -> int:
    return max(int((eval.log.time or 0.0) * 1000) for eval in evals)


def _get_evals_formatted_time(evals: List[Evaluation]) -> str:
    max_time = _get_evals_time_in_ms(evals)
    return f'{max_time} ms'


def _print_solution_outcome(
    solution: Solution,
    evals: List[Evaluation],
    timeLimit: int,
    console: rich.console.Console,
    verification: VerificationLevel = VerificationLevel.NONE,
) -> bool:
    bad_verdicts = set()
    for eval in evals:
        if eval.result.outcome != Outcome.ACCEPTED:
            bad_verdicts.add(eval.result.outcome)

    unmatched_bad_verdicts = set(
        v for v in bad_verdicts if not solution.outcome.match(v)
    )
    matched_bad_verdicts = bad_verdicts - unmatched_bad_verdicts

    if unmatched_bad_verdicts:
        console.print('[error]FAILED[/error]', end=' ')
    else:
        console.print('[success]OK[/success]', end=' ')

    console.print(f'Expected: {solution.outcome}', end='')

    if unmatched_bad_verdicts:
        unmatched_bad_verdicts_names = set(v.name for v in unmatched_bad_verdicts)
        console.print(f', got: {" ".join(unmatched_bad_verdicts_names)}', end='')

    console.print()
    evals_time = _get_evals_time_in_ms(evals)
    if (
        not (matched_bad_verdicts - {Outcome.TIME_LIMIT_EXCEEDED})
        and verification.value >= VerificationLevel.FULL.value
        and evals_time > timeLimit
        and evals_time < timeLimit * 2
    ):
        console.print(
            '[yellow]WARNING[/yellow] The solution still passed in double TL.'
        )
    console.print(f'Time: {_get_evals_formatted_time(evals)}')
    return len(unmatched_bad_verdicts) == 0


def print_detailed_run_report(
    evals_per_solution: List[Dict[str, List[Evaluation]]],
    console: rich.console.Console,
    verification: environment.VerificationParam,
):
    pkg = package.find_problem_package_or_die()

    for group in pkg.testcases:
        console.print(f'[bold][status]{group.name}[/status][/bold]')

        table = rich.table.Table()

        for solution in pkg.solutions:
            table.add_column(f'[item]{solution.path}[/item]', justify='full')

        row_to_renderables: Dict[int, List[Any]] = collections.defaultdict(list)
        time_summary_row: List[str] = []

        for evals_per_group in evals_per_solution:
            evals = evals_per_group[group.name]
            max_time = _get_evals_formatted_time(evals)
            for i, eval in enumerate(evals):
                verdict = _get_testcase_markup_verdict(eval)
                time = _get_evals_formatted_time([eval])
                if time == max_time:
                    time = f'[underline][bold]{time}[/bold][/underline]'
                row_to_renderables[i].append(f'{verdict} {time}')

            time_summary_row.append('  ' + _get_evals_formatted_time(evals))

        for _, verdicts in sorted(row_to_renderables.items()):
            table.add_row(*verdicts)
        if len(row_to_renderables) > 1:
            # Only print footer row if there's more than one testcase.
            table.add_section()
            table.add_row(*time_summary_row)

        console.print(table)
    console.print()


def print_run_report(
    evals_per_solution: List[Dict[str, List[Evaluation]]],
    console: rich.console.Console,
    verification: environment.VerificationParam,
    detailed: bool = False,
) -> bool:
    pkg = package.find_problem_package_or_die()

    assert len(pkg.solutions) == len(evals_per_solution)

    if detailed:
        print_detailed_run_report(evals_per_solution, console, verification)

    ok = True
    for s, (solution, evals_per_group) in enumerate(
        zip(pkg.solutions, evals_per_solution)
    ):
        if not evals_per_group:
            continue
        solution_testdir = package.get_problem_runs_dir() / f'{s}'
        console.print(f'[item]{solution.path}[/item]', end=' ')
        console.print(f'({solution_testdir})')

        all_evals = []
        for group, evals in evals_per_group.items():
            all_evals.extend(evals)
            if detailed:
                continue
            console.print(f'[bold][status]{group}[/status][/bold]', end=' ')
            console.print(f'({_get_evals_formatted_time(evals)})', end=' ')
            for i, eval in enumerate(evals):
                console.print(f'{i}/', end='')
                console.print(_get_testcase_markup_verdict(eval), end=' ')
            console.print()

        cur_ok = _print_solution_outcome(
            solution,
            all_evals,
            pkg.timeLimit,
            console,
            verification=VerificationLevel(verification),
        )
        ok = ok and cur_ok
        console.print()

    return ok



import pathlib
from typing import Optional

from pydantic import BaseModel


class Preset(BaseModel):
    # Name of the preset.
    name: str

    # Path to the environment file that will be installed with this preset.
    # When copied to the box environment, the environment will be named `name`.
    env: Optional[pathlib.Path] = None

    # Path to the contest preset directory, relative to the preset directory.
    problem: Optional[pathlib.Path] = None

    # Path to the problem preset directory, relative to the preset directory.
    contest: Optional[pathlib.Path] = None



import pathlib

import pytest

from robox.box.environment import VerificationLevel
from robox.box.generators import (
    generate_outputs_for_testcases,
    generate_testcases,
)
from robox.box.solutions import run_solutions
from robox.grading.steps import Outcome


@pytest.mark.test_pkg('box1')
def test_solutions(pkg_from_testdata: pathlib.Path):
    generate_testcases()
    generate_outputs_for_testcases()

    res = run_solutions(verification=VerificationLevel.FULL)

    # First solution should pass all tests.
    assert all(chk.result.outcome == Outcome.ACCEPTED for chk in res[0]['gen1'])
    # 25 test should be WA for the second solution.
    assert res[1]['gen1'][3].result.outcome == Outcome.WRONG_ANSWER
    # Runtime error for third solution.
    assert all(chk.result.outcome == Outcome.RUNTIME_ERROR for chk in res[2]['gen1'])
    # 1e9 test should be TLE for the fourth solution (soft TLE)
    assert res[3]['gen1'][4].result.outcome == Outcome.TIME_LIMIT_EXCEEDED
    # no TLE outcome should be WA (soft TLE)
    assert res[4]['gen1'][4].result.no_tle_outcome == Outcome.WRONG_ANSWER
    # hard TLE
    assert res[5]['gen1'][4].result.outcome in [
        Outcome.RUNTIME_ERROR,
        Outcome.TIME_LIMIT_EXCEEDED,
    ]



import os
import pathlib
import tempfile
from collections.abc import Iterator

import pytest

from robox.testing_utils import get_testdata_path


@pytest.fixture
def testdata_path() -> pathlib.Path:
    return get_testdata_path()


@pytest.fixture
def cleandir() -> Iterator[pathlib.Path]:
    with tempfile.TemporaryDirectory() as newpath:
        abspath = pathlib.Path(newpath).absolute()
        old_cwd = pathlib.Path.cwd()
        os.chdir(newpath)
        try:
            yield abspath
        finally:
            os.chdir(str(old_cwd))



from __future__ import annotations

import logging
import os
import pathlib
import resource
import shutil
import stat
import subprocess
import tempfile
from functools import partial
from sys import platform
from time import monotonic
from typing import IO, List, Optional, Union

import gevent

from robox.grading.judge import sandbox
from robox.grading.judge.cacher import FileCacher
from robox.grading.judge.sandbox import (
    SandboxBase,
    SandboxParams,
    wait_without_std,
)

logger = logging.getLogger(__name__)


class StupidSandbox(SandboxBase):
    """A stupid sandbox implementation. It has very few features and
    is not secure against things like box escaping and fork
    bombs. Yet, it is very portable and has no dependencies, so it's
    very useful for testing. Using in real contests is strongly
    discouraged.

    """

    exec_num: int
    popen: Optional[subprocess.Popen]
    popen_time: Optional[float]
    exec_time: Optional[float]

    def __init__(
        self,
        file_cacher: Optional[FileCacher] = None,
        name: Optional[str] = None,
        temp_dir: Optional[pathlib.Path] = None,
        params: Optional[SandboxParams] = None,
    ):
        """Initialization.

        For arguments documentation, see SandboxBase.__init__.

        """
        if not temp_dir:
            temp_dir = pathlib.Path(tempfile.gettempdir())
        SandboxBase.__init__(self, file_cacher, name, temp_dir, params)

        # Make box directory
        self._path = pathlib.Path(
            tempfile.mkdtemp(dir=str(self.temp_dir), prefix='rbx-%s-' % (self.name))
        )
        self.initialize()

        self.exec_num = -1
        self.popen = None
        self.popen_time = None
        self.exec_time = None

        logger.debug("Sandbox in `%s' created, using stupid box.", self._path)

        # Box parameters
        self.chdir = self._path

    def initialize(self):
        self._path.mkdir(parents=True, exist_ok=True)

    def get_root_path(self) -> pathlib.Path:
        """Return the toplevel path of the sandbox.

        return (Path): the root path.

        """
        return self._path

    # TODO - It returns wall clock time, because I have no way to
    # check CPU time (libev doesn't have wait4() support)
    def get_execution_time(self) -> Optional[float]:
        """Return the time spent in the sandbox.

        return (float): time spent in the sandbox.

        """
        return self.get_execution_wall_clock_time()

    # TODO - It returns the best known approximation of wall clock
    # time; unfortunately I have no way to compute wall clock time
    # just after the child terminates, because I have no guarantee
    # about how the control will come back to this class
    def get_execution_wall_clock_time(self) -> Optional[float]:
        """Return the total time from the start of the sandbox to the
        conclusion of the task.

        return (float): total time the sandbox was alive.

        """
        if self.exec_time:
            return self.exec_time
        if self.popen_time:
            self.exec_time = monotonic() - self.popen_time
            return self.exec_time
        return None

    def use_soft_timeout(self) -> bool:
        if platform == 'darwin':
            return False
        return True

    # TODO - It always returns None, since I have no way to check
    # memory usage (libev doesn't have wait4() support)
    def get_memory_used(self) -> Optional[int]:
        """Return the memory used by the sandbox.

        return (int): memory used by the sandbox (in bytes).

        """
        return None

    def get_killing_signal(self) -> int:
        """Return the signal that killed the sandboxed process.

        return (int): offending signal, or 0.

        """
        assert self.popen is not None
        if self.popen.returncode < 0:
            return -self.popen.returncode
        return 0

    # This sandbox only discriminates between processes terminating
    # properly or being killed with a signal; all other exceptional
    # conditions (RAM or CPU limitations, ...) result in some signal
    # being delivered to the process
    def get_exit_status(self) -> str:
        """Get information about how the sandbox terminated.

        return (string): the main reason why the sandbox terminated.

        """
        assert self.popen
        if self.popen.returncode >= 0:
            return self.EXIT_OK
        else:
            if -self.popen.returncode == 24:
                return self.EXIT_TIMEOUT
            return self.EXIT_SIGNAL

    def get_exit_code(self) -> int:
        """Return the exit code of the sandboxed process.

        return (float): exitcode, or 0.

        """
        assert self.popen
        return self.popen.returncode

    def get_human_exit_description(self) -> str:
        """Get the status of the sandbox and return a human-readable
        string describing it.

        return (string): human-readable explaination of why the
                         sandbox terminated.

        """
        status = self.get_exit_status()
        if status == self.EXIT_OK:
            return (
                'Execution successfully finished (with exit code %d)'
                % self.get_exit_code()
            )
        elif status == self.EXIT_SIGNAL:
            return 'Execution killed with signal %s' % self.get_killing_signal()
        return ''

    def hydrate_logs(self):
        return

    def _popen(
        self,
        command: List[str],
        stdin: Optional[IO[bytes] | int] = None,
        stdout: Optional[IO[bytes] | int] = None,
        stderr: Optional[IO[bytes] | int] = None,
        preexec_fn=None,
        close_fds: bool = True,
    ) -> subprocess.Popen:
        """Execute the given command in the sandbox using
        subprocess.Popen, assigning the corresponding standard file
        descriptors.

        command ([string]): executable filename and arguments of the
            command.
        stdin (file|None): a file descriptor/object or None.
        stdout (file|None): a file descriptor/object or None.
        stderr (file|None): a file descriptor/object or None.
        preexec_fn (function|None): to be called just before execve()
            or None.
        close_fds (bool): close all file descriptor before executing.

        return (object): popen object.

        """
        self.exec_time = None
        self.exec_num += 1

        logger.debug(
            "Executing program in sandbox with command: `%s'.", ' '.join(command)
        )
        with open(
            self.relative_path(self.cmd_file), 'at', encoding='utf-8'
        ) as commands:
            commands.write('%s\n' % command)
        try:
            p = subprocess.Popen(
                command,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                preexec_fn=preexec_fn,
                close_fds=close_fds,
            )
        except OSError:
            logger.critical(
                'Failed to execute program in sandbox ' "with command: `%s'.",
                ' '.join(command),
                exc_info=True,
            )
            raise

        return p

    def execute_without_std(
        self, command: List[str], wait: bool = False
    ) -> Union[bool, subprocess.Popen]:
        """Execute the given command in the sandbox using
        subprocess.Popen and discarding standard input, output and
        error. More specifically, the standard input gets closed just
        after the execution has started; standard output and error are
        read until the end, in a way that prevents the execution from
        being blocked because of insufficient buffering.

        command ([string]): executable filename and arguments of the
            command.

        return (bool): True if the sandbox didn't report errors
            (caused by the sandbox itself), False otherwise

        """

        def preexec_fn(self: 'StupidSandbox'):
            """Set limits for the child process."""
            if self.chdir:
                os.chdir(self.chdir)

            if platform == 'darwin':
                # These RLimits do not work properly on macOS.
                return

            # TODO - We're not checking that setrlimit() returns
            # successfully (they may try to set to higher limits than
            # allowed to); anyway, this is just for testing
            # environment, not for real contests, so who cares.
            if self.params.timeout:
                rlimit_cpu = self.params.timeout
                if self.params.extra_timeout:
                    rlimit_cpu += self.params.extra_timeout
                rlimit_cpu = int((rlimit_cpu + 999) // 1000)
                resource.setrlimit(resource.RLIMIT_CPU, (rlimit_cpu, rlimit_cpu))

            if self.params.address_space:
                rlimit_data = self.params.address_space * 1024 * 1024
                resource.setrlimit(resource.RLIMIT_DATA, (rlimit_data, rlimit_data))

            if self.params.stack_space:
                rlimit_stack = self.params.stack_space * 1024 * 1024
                resource.setrlimit(resource.RLIMIT_STACK, (rlimit_stack, rlimit_stack))

            # TODO - Doesn't work as expected
            # resource.setrlimit(resource.RLIMIT_NPROC, (1, 1))

        # Setup std*** redirection
        if self.params.stdin_file:
            stdin_fd = os.open(
                os.path.join(self._path, self.params.stdin_file), os.O_RDONLY
            )
        else:
            stdin_fd = subprocess.PIPE
        if self.params.stdout_file:
            stdout_fd = os.open(
                os.path.join(self._path, self.params.stdout_file),
                os.O_WRONLY | os.O_TRUNC | os.O_CREAT,
                stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR,
            )
        else:
            stdout_fd = subprocess.PIPE
        if self.params.stderr_file:
            if self.params.stderr_file == sandbox.MERGE_STDERR:
                stderr_fd = subprocess.STDOUT
            else:
                stderr_fd = os.open(
                    os.path.join(self._path, self.params.stderr_file),
                    os.O_WRONLY | os.O_TRUNC | os.O_CREAT,
                    stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR,
                )
        else:
            stderr_fd = subprocess.PIPE

        # Note down execution time
        self.popen_time = monotonic()

        # Actually call the Popen
        self.popen = self._popen(
            command,
            stdin=stdin_fd,
            stdout=stdout_fd,
            stderr=stderr_fd,
            preexec_fn=partial(preexec_fn, self),
            close_fds=True,
        )

        # Close file descriptors passed to the child
        if self.params.stdin_file:
            os.close(stdin_fd)
        if self.params.stdout_file:
            os.close(stdout_fd)
        if self.params.stderr_file and self.params.stderr_file != sandbox.MERGE_STDERR:
            os.close(stderr_fd)

        if self.params.wallclock_timeout:
            # Kill the process after the wall clock time passed
            def timed_killer(timeout, popen):
                gevent.sleep(timeout)
                # TODO - Here we risk to kill some other process that gets
                # the same PID in the meantime; I don't know how to
                # properly solve this problem
                try:
                    popen.kill()
                except OSError:
                    # The process had died by itself
                    pass

            # Setup the killer
            full_wallclock_timeout = self.params.wallclock_timeout
            if self.params.extra_timeout:
                full_wallclock_timeout += self.params.extra_timeout
            gevent.spawn(timed_killer, full_wallclock_timeout / 1000, self.popen)

        # If the caller wants us to wait for completion, we also avoid
        # std*** to interfere with command. Otherwise we let the
        # caller handle these issues.
        if wait:
            with self.popen as p:
                # Ensure popen fds are closed.
                res = self.translate_box_exitcode(wait_without_std([p])[0])
                # Ensure exec time is computed.
                self.get_execution_wall_clock_time()
                return res
        else:
            return self.popen

    def translate_box_exitcode(self, _):
        """Translate the sandbox exit code to a boolean sandbox success.

        This sandbox never fails.

        """
        return True

    def cleanup(self, delete=False):
        """See Sandbox.cleanup()."""
        # This sandbox doesn't have any cleanup, but we might want to delete.
        if delete:
            logger.debug('Deleting sandbox in %s.', self._path)
            shutil.rmtree(str(self._path))



import pathlib
import shutil
from typing import Optional

import rich
import rich.prompt
import typer

from robox import console, utils
from robox.box.environment import get_environment_path
from robox.box.presets.schema import Preset
from robox.config import get_default_app_path

app = typer.Typer(no_args_is_help=True)


def find_preset_yaml(root: pathlib.Path = pathlib.Path()) -> Optional[pathlib.Path]:
    found = root / 'preset.rbx.yml'
    if found.exists():
        return found
    return None


def get_preset_yaml(root: pathlib.Path = pathlib.Path()) -> Preset:
    found = find_preset_yaml(root)
    if not found:
        console.console.print(
            f'[error]preset.rbx.yml not found in {root.absolute()}[/error]'
        )
        raise typer.Exit(1)
    return utils.model_from_yaml(Preset, found.read_text())


def get_preset_installation_path(name: str) -> pathlib.Path:
    return utils.get_app_path() / 'presets' / name


def _try_installing_from_resources(name: str) -> bool:
    rsrc_preset_path = get_default_app_path() / 'presets' / name
    if not rsrc_preset_path.exists():
        return False
    yaml_path = rsrc_preset_path / 'preset.rbx.yml'
    if not yaml_path.is_file():
        return False
    console.console.print(f'Installing preset [item]{name}[/item] from resources...')
    _install(rsrc_preset_path, force=True)
    return True


def get_installed_preset(name: str) -> Preset:
    installation_path = get_preset_installation_path(name) / 'preset.rbx.yml'
    if not installation_path.is_file():
        if not _try_installing_from_resources(name):
            console.console.print(
                f'[error]Preset [item]{name}[/item] is not installed.[/error]'
            )
            raise typer.Exit(1)
    return utils.model_from_yaml(Preset, installation_path.read_text())


def _install(root: pathlib.Path = pathlib.Path(), force: bool = False):
    preset = get_preset_yaml(root)

    console.console.print(f'Installing preset [item]{preset.name}[/item]...')

    if preset.env is not None:
        console.console.print('Copying environment file...')
        if get_environment_path(preset.name).exists():
            res = force or rich.prompt.Confirm.ask(
                f'Environment [item]{preset.name}[/item] already exists. Overwrite?',
                console=console.console,
            )
            if not res:
                raise typer.Exit(1)

        shutil.copyfile(str(root / preset.env), get_environment_path(preset.name))

    console.console.print('Copying preset folder...')
    installation_path = get_preset_installation_path(preset.name)
    installation_path.parent.mkdir(parents=True, exist_ok=True)
    if installation_path.exists():
        res = force or rich.prompt.Confirm.ask(
            f'Preset [item]{preset.name}[/item] is already installed. Overwrite?',
            console=console.console,
        )
        if not res:
            raise typer.Exit(1)
    shutil.rmtree(str(installation_path), ignore_errors=True)
    shutil.copytree(str(root), str(installation_path))
    shutil.rmtree(str(installation_path / 'build'), ignore_errors=True)
    shutil.rmtree(str(installation_path / '.box'), ignore_errors=True)


@app.command('install', help='Install preset from current directory.')
def install():
    _install()


@app.callback()
def callback():
    pass



from gevent import monkey; monkey.patch_all()

import pathlib
import shutil
from typing import Annotated, Optional

import rich
import rich.prompt
import typer

from robox import annotations, config, console, utils
from robox.box import (
    builder,
    creation,
    download,
    environment,
    package,
    presets,
    stresses,
)
from robox.box.contest import main as contest
from robox.box.environment import VerificationLevel, get_environment_path
from robox.box.packaging import main as packaging
from robox.box.solutions import print_run_report, run_solutions
from robox.box.statements import build_statements

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)
app.add_typer(
    build_statements.app,
    name='statements, st',
    cls=annotations.AliasGroup,
    help='Manage statements.',
)
app.add_typer(
    download.app,
    name='download',
    cls=annotations.AliasGroup,
    help='Download an asset from supported repositories.',
)
app.add_typer(
    presets.app, name='presets', cls=annotations.AliasGroup, help='Manage presets.'
)
app.add_typer(
    packaging.app,
    name='package, pkg',
    cls=annotations.AliasGroup,
    help='Build problem packages.',
)
app.add_typer(
    contest.app, name='contest', cls=annotations.AliasGroup, help='Contest management.'
)


@app.command('edit, e', help='Open problem.rbx.yml in your default editor.')
def edit():
    console.console.print('Opening problem definition in editor...')
    # Call this function just to raise exception in case we're no in
    # a problem package.
    package.find_problem()
    config.open_editor(package.find_problem_yaml() or pathlib.Path())


@app.command('build, b', help='Build all tests for the problem.')
def build(verification: environment.VerificationParam):
    builder.build(verification=verification)


@app.command('verify, v', help='Build and verify all the tests for the problem.')
def verify(verification: environment.VerificationParam):
    if not builder.verify(verification=verification):
        console.console.print('[error]Verification failed, check the report.[/error]')


@app.command('run, r', help='Build and run solution(s).')
def run(
    verification: environment.VerificationParam,
    solution: Annotated[
        Optional[str],
        typer.Argument(
            help='Path to solution to run. If not specified, will run all solutions.'
        ),
    ] = None,
    detailed: bool = typer.Option(
        False,
        '--detailed',
        '-d',
        help='Whether to print a detailed view of the tests using tables.',
    ),
):
    builder.build(verification=verification)

    with utils.StatusProgress('Running solutions...') as s:
        tracked_solutions = None
        if solution:
            tracked_solutions = {solution}
        evals_per_solution = run_solutions(
            s,
            tracked_solutions=tracked_solutions,
        )

    console.console.print()
    console.console.rule('[status]Run report[/status]', style='status')
    print_run_report(
        evals_per_solution, console.console, verification, detailed=detailed
    )


@app.command('create, c', help='Create a new problem package.')
def create(
    name: str,
    preset: Annotated[
        Optional[str], typer.Option(help='Preset to use when creating the problem.')
    ] = None,
):
    creation.create(name, preset=preset)


@app.command('stress', help='Run a stress test.')
def stress(
    name: str,
    timeout: Annotated[
        int, typer.Option(help='For how many seconds to run the stress test.')
    ] = 10,
    findings: Annotated[
        int, typer.Option(help='How many breaking tests to look for.')
    ] = 1,
):
    # Do not verify built package.
    builder.build(verification=VerificationLevel.NONE.value)

    with utils.StatusProgress('Running stress...') as s:
        finding_list = stresses.run_stress(
            name, timeout, findingsLimit=findings, progress=s
        )

    stresses.print_stress_report(finding_list)

    if not finding_list:
        return

    # Add found tests.
    res = rich.prompt.Confirm.ask(
        'Do you want to add the tests that were found to a test group?',
        console=console.console,
    )
    if not res:
        return
    testgroup = None
    while testgroup is None or testgroup:
        testgroup = rich.prompt.Prompt.ask(
            'Enter the name of the test group, or empty to cancel',
            console=console.console,
        )
        if not testgroup:
            break
        try:
            testgroup = package.get_testgroup(testgroup)
            # Reassign mutable object before saving.
            testgroup.generators = testgroup.generators + [
                f.generator for f in finding_list
            ]
            package.save_package()
            console.console.print(
                f'Added [item]{len(finding_list)}[/item] tests to test group [item]{testgroup.name}[/item].'
            )
        except typer.Exit:
            continue
        break


@app.command('environment, env', help='Set or show the current box environment.')
def environment_command(env: Annotated[Optional[str], typer.Argument()] = None):
    if env is None:
        cfg = config.get_config()
        console.console.print(f'Current environment: [item]{cfg.boxEnvironment}[/item]')
        return
    if not get_environment_path(env).is_file():
        console.console.print(
            f'[error]Environment [item]{env}[/item] does not exist.[/error]'
        )
        raise typer.Exit(1)

    cfg = config.get_config()
    if env == cfg.boxEnvironment:
        console.console.print(
            f'Environment is already set to [item]{env}[/item].',
        )
        return
    console.console.print(
        f'Changing box environment from [item]{cfg.boxEnvironment}[/item] to [item]{env}[/item]...'
    )
    cfg.boxEnvironment = env
    config.save_config(cfg)

    # Also clear cache when changing environments.
    clear()


@app.command('clear, clean', help='Clears cache and build directories.')
def clear():
    console.console.print('Cleaning cache and build directories...')
    shutil.rmtree('.box', ignore_errors=True)
    shutil.rmtree('build', ignore_errors=True)


@app.callback()
def callback():
    pass



import pathlib
import shutil
import typing
from math import fabs
from typing import List, Literal

import typer
from pydantic import BaseModel

from robox import console
from robox.box import package
from robox.box.environment import get_extension_or_default
from robox.box.packaging.packager import BasePackager, BuiltStatement
from robox.box.statements.schema import Statement
from robox.config import get_default_app_path, get_testlib

BocaLanguage = Literal['c', 'cpp', 'cc', 'kt', 'java', 'py2', 'py3']

_MAX_REP_TIME = (
    7  # TL to allow for additional rounding reps should be < _MAX_REP_TIME in seconds
)
_MAX_REPS = 10  # Maximum number of reps to add
_MAX_REP_ERROR = 0.2  # 20% error allowed in time limit when adding reps


class BocaExtension(BaseModel):
    stdcpp: str = 'c++17'
    languages: List[BocaLanguage] = list(typing.get_args(BocaLanguage))
    maximumTimeError: float = _MAX_REP_ERROR


def test_time(time):
    return max(1, round(time))


class BocaPackager(BasePackager):
    def _get_main_statement(self) -> Statement:
        pkg = package.find_problem_package_or_die()

        if not pkg.statements:
            console.console.print('[error]No statements found.[/error]')
            raise typer.Exit(1)

        return pkg.statements[0]

    def _get_main_built_statement(
        self, built_statements: List[BuiltStatement]
    ) -> BuiltStatement:
        statement = self._get_main_statement()
        for built_statement in built_statements:
            if built_statement.statement == statement:
                return built_statement

        console.console.print(
            '[error]Main statement not found among built statements.[/error]'
        )
        raise typer.Exit(1)

    def _get_problem_name(self) -> str:
        pkg = package.find_problem_package_or_die()
        return pkg.name

    def _get_problem_info(self) -> str:
        pkg = package.find_problem_package_or_die()
        statement = self._get_main_statement()
        return (
            f'basename={pkg.name}\n'
            f'fullname={statement.title}\n'
            f'descfile={self._get_problem_name()}.pdf\n'
        )

    def _get_number_of_runs(self):
        pkg = package.find_problem_package_or_die()
        extension = get_extension_or_default('boca', BocaExtension)
        time = pkg.timeLimit / 1000  # convert to seconds

        if time >= _MAX_REP_TIME:
            return 1

        def rounding_error(time):
            return fabs(time - test_time(time))

        def error_percentage(time, runs):
            return rounding_error(time * runs) / (time * runs)

        if error_percentage(time, 1) < 1e-6:
            return 1

        for i in range(1, _MAX_REPS + 1):
            if error_percentage(time, i) <= extension.maximumTimeError:
                console.console.print(
                    f'[warning]Using {i} run(s) to define integer TL for BOCA (original TL is {pkg.timeLimit}ms, '
                    f'new TL is {test_time(time * i) * 1000}ms).[/warning]'
                )
                return i

        percent_str = f'{round(extension.maximumTimeError * 100)}%'
        console.console.print(
            f'[error]Error while defining limits for problem [item]{pkg.name}[/item].[/error]'
        )
        console.console.print(
            f'[error]Introducing an error of less than {percent_str} in the TL in less than '
            f'{_MAX_REPS} runs is not possible.[/error]'
        )
        console.console.print(
            f'[error]Original TL is {pkg.timeLimit}ms, please review it.[/error]'
        )
        raise typer.Exit(1)

    def _get_limits(self, no_of_runs: int) -> str:
        pkg = package.find_problem_package_or_die()
        return (
            '#!/bin/bash\n'
            f'echo {test_time(pkg.timeLimit / 1000 * no_of_runs)}\n'
            f'echo {no_of_runs}\n'
            f'echo {pkg.memoryLimit}\n'
            f'echo 4096\n'
            f'exit 0\n'
        )

    def _get_compare(self) -> str:
        compare_path = get_default_app_path() / 'packagers' / 'boca' / 'compare'
        if not compare_path.exists():
            console.console.print(
                '[error]BOCA template compare script not found.[/error]'
            )
            raise typer.Exit(1)
        return compare_path.read_text()

    def _get_checker(self) -> str:
        checker_path = get_default_app_path() / 'packagers' / 'boca' / 'checker.sh'
        if not checker_path.exists():
            console.console.print(
                '[error]BOCA template checker script not found.[/error]'
            )
            raise typer.Exit(1)
        checker_text = checker_path.read_text()
        testlib = get_testlib().read_text()
        checker = package.get_checker().path.read_text()
        return checker_text.replace('{{testlib_content}}', testlib).replace(
            '{{checker_content}}', checker
        )

    def _get_compile(self, language: BocaLanguage) -> str:
        extension = get_extension_or_default('boca', BocaExtension)

        compile_path = (
            get_default_app_path() / 'packagers' / 'boca' / 'compile' / language
        )
        if not compile_path.is_file():
            console.console.print(
                f'[error]Compile script for language [item]{language}[/item] not found.[/error]'
            )
            raise typer.Exit(1)

        compile_text = compile_path.read_text()

        assert 'umask 0022' in compile_text
        compile_text = compile_text.replace(
            'umask 0022', 'umask 0022\n\n' + self._get_checker()
        )

        if language == 'cc':
            compile_text = compile_text.replace(
                '{{stdcpp}}', f'-std={extension.stdcpp}'
            )
        return compile_text

    def name(self) -> str:
        return 'boca'

    def package(
        self,
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltStatement],
    ) -> pathlib.Path:
        extension = get_extension_or_default('boca', BocaExtension)
        no_of_runs = self._get_number_of_runs()

        # Prepare limits
        limits_path = into_path / 'limits'
        limits_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            (limits_path / language).write_text(self._get_limits(no_of_runs))

        # Prepare compare
        compare_path = into_path / 'compare'
        compare_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            (compare_path / language).write_text(self._get_compare())

        # Prepare run
        run_path = into_path / 'run'
        run_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            run_orig_path = (
                get_default_app_path() / 'packagers' / 'boca' / 'run' / language
            )
            if not run_orig_path.is_file():
                console.console.print(
                    f'[error]Run script for language [item]{language}[/item] not found.[/error]'
                )
                raise typer.Exit(1)
            shutil.copyfile(run_orig_path, run_path / language)

        # Prepare compile.
        compile_path = into_path / 'compile'
        compile_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            (compile_path / language).write_text(self._get_compile(language))

        # Prepare tests
        tests_path = into_path / 'tests'
        tests_path.mkdir(parents=True, exist_ok=True)
        for language in extension.languages:
            (tests_path / language).write_text('exit 0\n')

        # Problem statement
        description_path = into_path / 'description'
        description_path.mkdir(parents=True, exist_ok=True)
        (description_path / 'problem.info').write_text(self._get_problem_info())
        shutil.copyfile(
            self._get_main_built_statement(built_statements).path,
            (description_path / self._get_problem_name()).with_suffix('.pdf'),
        )

        # Prepare IO
        inputs_path = into_path / 'input'
        inputs_path.mkdir(parents=True, exist_ok=True)
        outputs_path = into_path / 'output'
        outputs_path.mkdir(parents=True, exist_ok=True)

        testcases = self.get_flattened_built_testcases()
        for i, testcase in enumerate(testcases):
            shutil.copyfile(testcase.inputPath, inputs_path / f'{i+1:03d}')
            if testcase.outputPath is not None:
                shutil.copyfile(testcase.outputPath, outputs_path / f'{i+1:03d}')
            else:
                (outputs_path / f'{i+1:03d}').touch()

        # Zip all.
        shutil.make_archive(
            str(build_path / self._get_problem_name()), 'zip', into_path
        )

        return (build_path / self._get_problem_name()).with_suffix('.zip')



import pathlib
import random
import shlex
import time
from typing import List, Optional, Union

import typer
from pydantic import BaseModel

from robox import console
from robox.box import checkers, package
from robox.box.code import compile_item, run_item
from robox.box.schema import GeneratorCall, Testcase
from robox.box.solutions import compile_solutions, get_outcome_style_verdict
from robox.grading.steps import (
    CheckerResult,
    DigestOrDest,
    DigestOrSource,
)
from robox.utils import StatusProgress

StressArg = Union[str, 'RandomInt', 'RandomHex', List['StressArg']]


class StressFinding(BaseModel):
    generator: GeneratorCall
    solution: pathlib.Path
    result: CheckerResult


class RandomInt:
    def __init__(self, min: int, max: int):
        self.min = min
        self.max = max

    def get(self) -> int:
        return random.randint(self.min, self.max)


class RandomHex:
    len: int

    def __init__(self, len: int):
        self.len = len

    def get(self) -> str:
        return ''.join(random.choice('0123456789abcdef') for _ in range(self.len))


def _parse_random_choice(pattern: str) -> List[StressArg]:
    # TODO: Add escaping for |
    return [_parse_single_pattern(choice) for choice in pattern.split('|')]


def _parse_var(name: str) -> str:
    pkg = package.find_problem_package_or_die()

    if name not in pkg.vars:
        console.console.print(f'[error]Variable [item]{name}[/item] not found.[/error]')
        raise typer.Exit(1)
    return f'{pkg.expanded_vars[name]}'


def _parse_int(pattern: str) -> int:
    if pattern.startswith('<') and pattern.endswith('>'):
        return int(_parse_var(pattern[1:-1]))
    return int(pattern)


def _parse_random_int(pattern: str) -> RandomInt:
    min, max = pattern.split('..')
    return RandomInt(_parse_int(min), _parse_int(max))


def _parse_single_pattern(pattern: str) -> StressArg:
    if pattern.startswith('\\'):
        # Escape
        return pattern[1:]
    if pattern.startswith('<') and pattern.endswith('>'):
        return _parse_var(pattern[1:-1])
    if pattern.startswith('[') and pattern.endswith(']'):
        # Random range
        return _parse_random_int(pattern[1:-1])
    if pattern.startswith('(') and pattern.endswith(')'):
        return _parse_random_choice(pattern[1:-1])
    if pattern == '@':
        return RandomHex(len=8)
    return pattern


def parse_generator_pattern(args: str) -> List[StressArg]:
    return [_parse_single_pattern(arg) for arg in shlex.split(args)]


def _expand_single_arg(arg: StressArg) -> str:
    if isinstance(arg, RandomInt):
        return str(arg.get())
    if isinstance(arg, RandomHex):
        return arg.get()
    if isinstance(arg, list):
        return _expand_single_arg(random.choice(arg))
    return str(arg)


def expand_stress_args(pattern: List[StressArg]) -> List[str]:
    return [_expand_single_arg(arg) for arg in pattern]


def run_stress(
    name: str,
    timeoutInSeconds: int,
    findingsLimit: int = 1,
    progress: Optional[StatusProgress] = None,
) -> List[StressFinding]:
    stress = package.get_stress(name)

    call = stress.generator
    generator = package.get_generator(call.name)
    main_solution = package.get_main_solution()
    solutions = [package.get_solution(solutions) for solutions in stress.solutions]
    solutions = [main_solution] + solutions

    try:
        generator_digest = compile_item(generator)
    except:
        console.console.print(
            f'[error]Failed compiling generator [item]{generator.name}[/item].[/error]'
        )
        raise
    checker_digest = checkers.compile_checker()
    solutions_digest = compile_solutions(
        tracked_solutions=set(
            str(solution.path) for solution in solutions if solution is not None
        )
    )

    startTime = time.monotonic()
    parsed_args = parse_generator_pattern(call.args or '')
    runs_dir = package.get_problem_runs_dir()

    findings = []

    while len(findings) < findingsLimit:
        if time.monotonic() - startTime > timeoutInSeconds:
            break

        if progress:
            seconds = timeoutInSeconds - int(time.monotonic() - startTime)
            progress.update(
                f'Stress testing: found [item]{len(findings)}[/item] tests, '
                f'[item]{seconds}[/item] second(s) remaining...'
            )

        expanded_args = expand_stress_args(parsed_args)
        expanded_args_str = ' '.join(expanded_args)

        input_path = runs_dir / '.stress' / 'input'
        input_path.parent.mkdir(parents=True, exist_ok=True)

        generation_log = run_item(
            generator,
            DigestOrSource.create(generator_digest),
            stdout=DigestOrDest.create(input_path),
            extra_args=expanded_args_str or None,
        )
        if not generation_log or generation_log.exitcode != 0:
            console.console.print(
                f'Failed generating test for stress test [item]{name}[/item] with args [info]{expanded_args}[/info]',
                style='error',
            )
            raise typer.Exit(1)

        expected_output_path = runs_dir / '.stress' / 'output'
        for i, solution in enumerate(solutions):
            if solution is None:
                continue
            output_path = input_path.with_stem(f'{i}').with_suffix('.out')
            if i == 0:
                # This is the main solution.
                expected_output_path = output_path

            run_log = run_item(
                solution,
                DigestOrSource.create(solutions_digest[solution.path]),
                stdin=DigestOrSource.create(input_path),
                stdout=DigestOrDest.create(output_path),
            )

            checker_result = checkers.check(
                checker_digest,
                run_log,
                Testcase(inputPath=input_path, outputPath=expected_output_path),
                program_output=output_path,
            )

            if not stress.outcome.match(checker_result.outcome):
                continue

            if progress:
                console.console.print(
                    f'[error]FINDING[/error] Generator args are "[status]{generator.name} {expanded_args_str}[/status]"'
                )

            findings.append(
                StressFinding(
                    generator=GeneratorCall(
                        name=generator.name, args=expanded_args_str
                    ),
                    solution=solution.path,
                    result=checker_result,
                )
            )

        # Be cooperative.
        time.sleep(0.001)

    return findings


def print_stress_report(findings: List[StressFinding]):
    console.console.rule('Stress test report', style='status')
    if not findings:
        console.console.print('[info]No stress test findings.[/info]')
        return

    for i, finding in enumerate(findings):
        console.console.print(f'[error]Finding {i + 1}[/error]')
        console.console.print(
            f'Generator: [status]{finding.generator.name} {finding.generator.args}[/status]'
        )
        console.console.print(f'Solution: [item]{finding.solution}[/item]')
        style = get_outcome_style_verdict(finding.result.outcome)
        console.console.print(
            f'Outcome: [{style}]{finding.result.outcome.name}[/{style}]'
        )
        console.console.print(f'Message: [status]{finding.result.message}[/status]')
        console.console.print()
